from nmigen import *
from nmigen.cli import main


from functools import reduce

from skidbuffer import RegisteredSkidBuffer



class ParseStack(Elaboratable):
    PUSH = 0
    POP  = 1
    MULTIPOP = 2
    INTROSPECT = 3
    READ_SP    = 4
    READ_TOS   = 5
    def __init__(self, width, depth):
        # Parameters
        self.width = width
        self.depth = depth

        # Control inputs
        self.command_port = Signal(3)
        # The commands are as follows:
        # 0: push
        # 1: pop
        # 2: multipop (number to pop in aux port)
        # 3: introspect (index in aux port)
        # 4: read stack pointer
        # 5: read top of stack

        # pop outputs the value of the popped item on the output data port

        # multipop outputs the value *at the top of the stack* after all the popping is done
        # on the output data port

        # read stack pointer outputs the stack pointer on the aux port

        self.strobe       = Signal(1) # when 1, execute the command

        # Data inputs
        self.in_data_port      = Signal(width)
        self.in_aux_port       = Signal(range(depth))
        self.in_valid          = Signal(1)

        # Control outputs
        self.valid_out      = Signal(1)
        self.internal_fault = Signal(1)

        # Data outputs
        self.out_data_port       = Signal(width)
        self.out_aux_port        = Signal(range(depth))
        self.out_data_valid      = Signal(1)

        self.mem = Memory(width=self.width, depth=self.depth)


    def elaborate(self, platform):
        m = Module()

        m.submodules.rport = rport = (self.mem).read_port()
        m.submodules.wport = wport = (self.mem).write_port()
        stack_pointer = Signal(range(self.depth), reset=0)


        # To represent the state of a k-item stack, we need an additional bit of state to
        # disambiguate the case where there are zero items on the stack and the case where
        # there is a single item pushed on the stack.

        # bottom_of_stack_valid = 0 implies zero items on the stack (which means s_p must be 0), and
        # popping generates an error

        # bottom_of_stack_valid = 1 implies that there is at least one item on the stack, and s_p can take
        # any value.

        # 0 items on stack (bottom_of_stack_valid = 0, stack_pointer = 0)
        # PUSH
        # 1 item on stack  (bottom_of_stack_valid = 1, stack_pointer = 0)
        # PUSH
        # 2 items on stack (bottom_of_stack_valid = 1, stack_pointer = 1)
        # PUSH
        # 3 items on stack (bottom_of_stack_valid = 1, stack_pointer = 2)

        # and for popping, it proceeds in reverse.

        bottom_of_stack_valid = Signal(1)


        multipop_left = Signal(range(self.depth))

        with m.FSM() as fsm:
            with m.State("EMPTY"):
                with m.If(self.strobe == 1):
                    with m.Switch(self.command_port):
                        with m.Case(self.PUSH):
                            with m.If(self.in_valid == 1):
                                m.d.comb += wport.addr.eq(stack_pointer)
                                m.d.comb += wport.data.eq(self.in_data_port)
                                m.d.comb += wport.en.eq(1)
                                m.next = "AT_LEAST_ONE_ITEM"
                        with m.Case(self.POP):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
                        with m.Case(self.MULTIPOP):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
                        with m.Case(self.INTROSPECT):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
                        with m.Case(self.READ_TOS):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
                        with m.Case(self.READ_SP):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"

            with m.State("AT_LEAST_ONE_ITEM"):
                with m.If(self.strobe == 1):
                    with m.Switch(self.command_port):
                        with m.Case(self.PUSH):
                            with m.If(self.in_valid == 1):
                                m.d.sync += stack_pointer.eq(stack_pointer + 1)
                                m.d.comb += wport.addr.eq(stack_pointer + 1)
                                m.d.comb += wport.data.eq(self.in_data_port)
                                m.d.comb += wport.en.eq(1)
                                m.next = "AT_LEAST_ONE_ITEM"
                        with m.Case(self.POP):
                            m.d.comb += rport.addr.eq(stack_pointer)
                            m.d.comb += self.out_data_port.eq(rport.data)

                            with m.If(stack_pointer == 0):
                                m.next = "EMPTY"
                            with m.Else():
                                m.next = "AT_LEAST_ONE_ITEM"
                                m.d.sync += stack_pointer.eq(stack_pointer - 1)
                        with m.Case(self.MULTIPOP):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
                        with m.Case(self.INTROSPECT):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
                        with m.Case(self.READ_TOS):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
                        with m.Case(self.READ_SP):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
            with m.State("FAULTED"):
                m.d.comb += self.internal_fault.eq(1)
                m.next = "FAULTED"





 




        return m


class DummyPlug(Elaboratable):

    #def __init__(self):



    def elaborate(self, platform):
        m = Module()

        m.submodules.stack = ParseStack(8,8)
        counter = Signal(8)
        m.d.sync += counter.eq(counter+1)

        with m.If(counter==1):
            m.d.comb += m.submodules.stack.strobe.eq(1)
            m.d.comb += m.submodules.stack.in_valid.eq(1)
            m.d.comb += m.submodules.stack.in_data_port.eq(42)


        with m.If(counter==2):
            m.d.comb += m.submodules.stack.command_port.eq(0)
            m.d.comb += m.submodules.stack.strobe.eq(1)
            m.d.comb += m.submodules.stack.in_valid.eq(1)
            m.d.comb += m.submodules.stack.in_data_port.eq(43)

        with m.If(counter==3):
            m.d.comb += m.submodules.stack.command_port.eq(0)
            m.d.comb += m.submodules.stack.strobe.eq(1)
            m.d.comb += m.submodules.stack.in_valid.eq(1)
            m.d.comb += m.submodules.stack.in_data_port.eq(44)

        with m.If(counter==4):
            m.d.comb += m.submodules.stack.command_port.eq(1)
            m.d.comb += m.submodules.stack.strobe.eq(1)

        with m.If(counter==5):
            m.d.comb += m.submodules.stack.command_port.eq(1)
            m.d.comb += m.submodules.stack.strobe.eq(1)


        with m.If(counter==6):
            m.d.comb += m.submodules.stack.command_port.eq(1)
            m.d.comb += m.submodules.stack.strobe.eq(1)
        with m.If(counter==7):
            m.d.comb += m.submodules.stack.command_port.eq(1)
            m.d.comb += m.submodules.stack.strobe.eq(1)
        return m



if __name__ == '__main__':
    baka =DummyPlug()
    main(baka)
    #platform.build(DummyPlug())
